// Four Elements III Contest Entry
// Author: Gints Vitols (e-mail: gints@cs.rtu.lv, GameDev.Net username: Advanced Bug)


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include <stdlib.h>
#include <math.h>


//---------------------------------- Misc -------------------------------------


double rnd()
{
  return double(rand()) / RAND_MAX;
}


//----------------------------------- Image --------------------------------


struct IMAGE
{
  HBITMAP handle;
  HDC device;
  HBITMAP old_bmp;
};


void create_image(IMAGE *image)
{
  BITMAPINFO bmi;
  ZeroMemory(&bmi, sizeof(BITMAPINFO));
  BITMAPINFOHEADER *bmih = &bmi.bmiHeader;
  bmih->biSize = sizeof(BITMAPINFOHEADER);
  bmih->biWidth = 256;
  bmih->biHeight = 256;
  bmih->biPlanes = 1;
  bmih->biBitCount = 16;
  bmih->biCompression = BI_RGB;
  void *bits;
  image->handle = CreateDIBSection(NULL, &bmi, DIB_RGB_COLORS, &bits, NULL, 0);
  image->device = CreateCompatibleDC(NULL);
  image->old_bmp = (HBITMAP)SelectObject(image->device, image->handle);
}


void destroy_image(IMAGE *image)
{
  SelectObject(image->device, image->old_bmp);
  DeleteObject(image->device);
  DeleteObject(image->handle);
}


IMAGE land_image, buffer;


//----------------------------------- Main Window ----------------------------------


HWND main_window;


void create_main_window(void)
{
  RECT rect;
  rect.left = 0;
  rect.top = 0;
  rect.right = 256;
  rect.bottom = 256;
  AdjustWindowRect(&rect, WS_OVERLAPPEDWINDOW, FALSE);
  main_window = CreateWindow("MainWndClass", "4E3 Contest Entry", 
                             WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
                             CW_USEDEFAULT, CW_USEDEFAULT,
                             rect.right, rect.bottom,
                             NULL, NULL, (HINSTANCE)GetModuleHandle(NULL), NULL);
  ShowWindow(main_window, SW_SHOWDEFAULT);
}


void redraw_main_window()
{
  HDC device = GetDC(main_window);
  BitBlt(device, 0, 0, 256, 256, buffer.device, 0, 0, SRCCOPY);
  ReleaseDC(main_window, device);
}


//----------------------------------- Particle ---------------------------------


const int NUM_PARTICLES = 256;
const int MAX_COUNTER = 256;
const double WIND_SPEED = -.5;
const double WIND_PULSE_FREQ = .005;


int time = 0;


struct PARTICLE
{
  double x, y;
  double vy;
  int counter;
  bool active;
};


PARTICLE particles[NUM_PARTICLES];


void create_inactive_particle(PARTICLE *p, int c)
{
  p->active = false;
  p->counter = c;
}


void create_active_particle(PARTICLE *p)
{
  p->active = true;
  p->counter = 0;
  p->x = 16 * (2 * rnd() - 1);
  p->y = 0;
  p->vy = exp( - pow(p->x / 12, 2)) * (.5 + .5 * rnd());
}


void move_particle(PARTICLE *p)
{
  double vx = WIND_SPEED * (1 - cos(WIND_PULSE_FREQ * time));
  p->x += vx;
  p->y += p->vy;
}


void update_particle(PARTICLE *p)
{
  p->counter++;
  if(p->counter == MAX_COUNTER)
    create_active_particle(p);
  else if(p->active)
    move_particle(p);
}


struct COLOR
{
  double limit, r, g, b;
};


COLOR colors[] = { 0.00, 0,   0,   0, 
                   0.10, 255, 255, 255, 
                   0.20, 255, 255, 0,
                   0.80, 255, 0,   0,
                   1.00, 0,   0,   0  };


COLORREF get_particle_color(PARTICLE *p)
{
  double i = double(p->counter) / MAX_COUNTER;
  double r, g, b;
  COLOR *c1 = &colors[0], *c2 = &colors[1];
  while(i > c2->limit)
    c1 = c2++;
  double i1 = (i - c1->limit) / (c2->limit - c1->limit);
  r = c1->r + (c2->r - c1->r) * i1;
  g = c1->g + (c2->g - c1->g) * i1;
  b = c1->b + (c2->b - c1->b) * i1;
  return RGB(r, g, b);
}


void render_particle(PARTICLE *p)
{
  COLORREF c = get_particle_color(p);
  int x = 128 + int(p->x);
  int y1 = 128 - int(p->y);
  int y2 = 128 + int(p->y / 2);

  SetPixel(buffer.device, x, y1, c);
  SetPixel(buffer.device, x + 1, y1, c);
  SetPixel(buffer.device, x, y1 + 1, c);
  SetPixel(buffer.device, x, y1 - 1, c);
  SetPixel(buffer.device, x - 1, y1, c);

  if(GetPixel(buffer.device, x, y2) == 0)
    SetPixel(buffer.device, x, y2, c);
  if(GetPixel(buffer.device, x - 1, y2) == 0)
    SetPixel(buffer.device, x - 1, y2, c);
  if(GetPixel(buffer.device, x + 1, y2) == 0)
    SetPixel(buffer.device, x + 1, y2, c);
}


//----------------------------------- Particles ------------------------------


void init_particles()
{
  for(int i = 0; i < NUM_PARTICLES; i++)
    create_inactive_particle(&particles[i], MAX_COUNTER * i / NUM_PARTICLES);
}


void update_particles()
{
  for(int i = 0; i < NUM_PARTICLES; i++)
    update_particle(&particles[i]);
}


void render_particles()
{
  for(int i = 0; i < NUM_PARTICLES; i++)
    render_particle(&particles[i]);
}


//------------------------------------ Land --------------------------------


COLORREF get_land_pixel()
{
  double r1 = 128, g1 = 128, b1 = 128;
  double r2 = 0, g2 = 64, b2 = 0;
  double r, g, b;
  double i = rnd();
  r = r1 + i * (r2 - r1);
  g = g1 + i * (g2 - g1);
  b = b1 + i * (b2 - b1);
  return RGB(r, g, b);
}


void init_land_image()
{
  PatBlt(land_image.device, 0, 0, 256, 256, BLACKNESS);
  for(int x = 0; x < 256; x++)
    for(int y = 0; y < 256; y++)
      if(sqrt(pow(x - 128, 2) + pow(2 * (y - 128), 2)) < 32)
        SetPixel(land_image.device, x, y, get_land_pixel());
}


//------------------------------------ Scene -------------------------------


void create_scene()
{
  create_image(&buffer);
  create_image(&land_image);
  init_land_image();
  init_particles();
}


void update_scene()
{
  BitBlt(buffer.device, 0, 0, 256, 256, land_image.device, 0, 0, SRCCOPY);
  update_particles();
  render_particles();  
  redraw_main_window();  
  time++;
}


void destroy_scene()
{
  destroy_image(&buffer);
  destroy_image(&land_image);
}


//------------------------------ Main -----------------------------------


LRESULT CALLBACK MainWndProc(HWND hwnd, UINT uMsg,
                             WPARAM wParam, LPARAM lParam)
{
  switch(uMsg)
  {
    case WM_DESTROY:
      PostQuitMessage(0);
      break;
  }
  return DefWindowProc(hwnd, uMsg, wParam, lParam);
}


void register_window_class()
{
  WNDCLASS wc;
  wc.style = 0;
  wc.lpfnWndProc = MainWndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = (HINSTANCE)GetModuleHandle(NULL);
  wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
  wc.lpszMenuName =  NULL;
  wc.lpszClassName = "MainWndClass";
  RegisterClass(&wc);
}


void do_message_loop()
{
  MSG msg;
  bool quit = false;
  DWORD old_time = timeGetTime();
  while(!quit)
  {
    while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
    {
      if(msg.message == WM_QUIT)
      {
        quit = true;
        break;
      }
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    DWORD new_time = timeGetTime();
    if((new_time - old_time) > 10)
    {
      old_time = new_time;
      update_scene();
    }
  }
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpszCmdLine, int nCmdShow)
{
  create_scene();
  register_window_class();
  create_main_window();
  do_message_loop();
  destroy_scene();
}
